# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from torchmetrics.regression.concordance import ConcordanceCorrCoef  # noqa: F401
from torchmetrics.regression.cosine_similarity import CosineSimilarity  # noqa: F401
from torchmetrics.regression.explained_variance import ExplainedVariance  # noqa: F401
from torchmetrics.regression.kl_divergence import KLDivergence  # noqa: F401
from torchmetrics.regression.log_mse import MeanSquaredLogError  # noqa: F401
from torchmetrics.regression.mae import MeanAbsoluteError  # noqa: F401
from torchmetrics.regression.mape import MeanAbsolutePercentageError  # noqa: F401
from torchmetrics.regression.mse import MeanSquaredError  # noqa: F401
from torchmetrics.regression.pearson import PearsonCorrCoef  # noqa: F401
from torchmetrics.regression.r2 import R2Score  # noqa: F401
from torchmetrics.regression.spearman import SpearmanCorrCoef  # noqa: F401
from torchmetrics.regression.symmetric_mape import SymmetricMeanAbsolutePercentageError  # noqa: F401
from torchmetrics.regression.tweedie_deviance import TweedieDevianceScore  # noqa: F401
from torchmetrics.regression.wmape import WeightedMeanAbsolutePercentageError  # noqa: F401
